#include <util.h>
#include <string.h>

opts_t
parse_opts (int argc, char **argv)
{

  opts_t opts = {
    .keystone = {.argc = 1,.argv = NULL},
    .outfile = NULL
  };

  int i;

  for (i = 1; i < argc; i++)
    {
      if (!strcmp (argv[i], "-o"))
	{
	  i++;
	  opts.outfile = argv[i];
	}
      if (!strcmp (argv[i], "--help"))
	{
	  printf ("usage: gsc module.cpp -I include_dir/ -o module.v\n");
	  printf ("options\n");
	  printf ("\t-o module.v,   default output file\n");
	  printf ("\t--help,        prints this\n");
	  printf ("\t--version,     prints gsc version's\n");
	  exit (1);
	}
      if (!strcmp (argv[i], "--version"))
	{
	  printf ("gsc (gnu systemc compiler) 1.0.1\n");
	  exit (1);
	}
      else
	{
	  opts.keystone.argc++;
	  opts.keystone.argv =
	    (char **) realloc (opts.keystone.argv,
			       opts.keystone.argc * sizeof (char *));
	  opts.keystone.argv[opts.keystone.argc - 1] = argv[i];
	}
    }


  return opts;
}

inline Agnode_t *
get_fschild (Agnode_t * node)
{
  Agedge_t *e = first_child (node);
  if (e)
    return e->head;
  return NULL;
}

inline Agnode_t *
get_scchild (Agnode_t * node)
{
  Agedge_t *e = agfstout (node->graph, node);
  do
    {
      e = agnxtout (node->graph, e);
    }
  while (e && e->head == get_fschild (node));

  if (e)
    return e->head;

  return NULL;
}

inline Agnode_t *
get_thchild (Agnode_t * node)
{
  Agedge_t *e = third_child (node);
  if (e)
    return e->head;
  return NULL;
}

inline Agnode_t *
get_fhchild (Agnode_t * node)
{
  Agedge_t *e = fourth_child (node);
  if (e)
    return e->head;
  return NULL;
}

inline char *
get_fsattr (Agnode_t * node)
{

  if (node == NULL)
    return NULL;

  char *str = node->attr[0];
  assert (str[0] == '{');
  assert (str[strlen (str) - 1] == '}');

  int i = 1;
  int length = strlen (str);
  int begin = 0, end = 0;
  for (i = 1; i < length; i++)
    {
      if (str[i] == ':')
	begin = i + 2;
      if (str[i] == '|')
	{
	  end = i - 1;
	  assert (end > begin);
	  char *s = (char *) strndup (&str[begin], end - begin);
	  //char *s = strndup (&str[begin], end - begin);
	  s[end - begin] = '\0';
	  return s;
	}
      if (str[i] == '}')
	{
	  end = i;
	  assert (end > begin);
	  char *s = (char *) strndup (&str[begin], end - begin);
	  s[end - begin] = '\0';
	  return s;
	}
    }

  return NULL;
}


inline int
get_attr (Agnode_t * node, char ***argv)
{

  char *str = node->attr[0];

  char **attrs, **p;

  assert (str[0] == '{');
  assert (str[strlen (str) - 1] == '}');

  char *c;
  int length = strlen (str);
  int i;
  int name = 1;
  int value = 1;

  int n = 1;
  int open = 1;
  for (i = 1; i < (length - 1); i++)
    {
      if ((str[i] == '|') && open)
	{
	  n++;
	  open = 0;
	}
      if (str[i] == ':')
	open = 1;
    }

  //printf("Found %d attributes\n", n);
  open = 0;
  p = attrs = (char **) malloc (sizeof (char *) * (2 * n));
  char *s;
  for (i = 1; i < (length - 1); i++)
    {
      if (str[i] == ':')
	{
	  assert (i > name);

	  if (name < value)
	    continue;

	  open = 1;

	  assert (name >= value);

	  s = (char *) strndup (&str[name], i - name);
	  s[(i - name)] = '\0';

	  *attrs = s;
	  attrs++;

	  value = i + 2;

	}

      if (str[i] == '}')
	{
	  printf ("hi\n");
	  break;
	}

      if ((str[i] == '|') || (i == (length - 2)))
	{
	  //char value_str[20];
	  open = 0;
	  int end = (i == (length - 2)) ? (i - value) + 1 : (i - value) - 1;

	  if (end <= 0)
	    {
	      //printf( "%d %d %d\n", length, value, i  );
	      //end = length - 1;
	      end = 2;
	    }

	  assert (end > 0);

	  s = (char *) strndup (&str[value], end);
	  s[end] = '\0';

	  *attrs = s;
	  attrs++;

	  if (!strcmp (s, "|") || !strcmp (s, "||"))
	    {
	      n--;
	      break;
	    }
	  name = i + 2;
	}

    }

  *argv = p;
  return 2 * n;

}

inline char *
get_namedattr (Agnode_t * node, char *name)
{

  if (node == NULL)
    return NULL;

  char **argv;
  int argc = get_attr (node, &argv);
  int i;
  char *str = NULL;

  for (i = 0; i < argc; i += 2)
    {
      assert (argv[i + 0]);
      assert (argv[i + 1]);
      if (!strcmp (name, argv[i + 0]))
	str = strdup (argv[i + 1]);
      free (argv[i + 0]);
      free (argv[i + 1]);
    }

  free (argv);
  //assert (str);
  return str;
}


static inline int
node_type (Agnode_t * root, char *type)
{
  char *s = get_fsattr (root);
  int result = !strcmp (type, s);
  free (s);
  return result;
}

void
__print_ast (Agraph_t * g, Agnode_t * root, int depth)
{
  Agedge_t *e;

  assert (g);
  assert (root);

  //if( depth == 2) return;

  printfin (stdout, depth, "(%s\n", root->name);
  char *s = get_fsattr (root);
  //char *s = get_namedattr (root,"name");
  //printfin (stdout, depth, " %s\n", s);
  free (s);

  char **argv;
  int argc = get_attr (root, &argv);
  int i;

  for (i = 0; i < argc; i += 2)
    {
      assert (argv[i + 0]);
      assert (argv[i + 1]);
      printfin (stdout, depth + 1, "%s = %s\n", argv[i + 0], argv[i + 1]);
      free (argv[i + 0]);
      free (argv[i + 1]);
    }

  free (argv);

  for_each_edge (e, root, g)
  {
    Agnode_t *head = e->head;
    __print_ast (g, head, depth + 1);
  }

  //indent (depth);
  printfin (stdout, depth, ")\n");

}

#define for_each_arg( ptr, type, args )  for( ptr = va_arg( args, type ) ; ptr ; ptr = va_arg( args, type ) )

Agnode_t *
resolve_path (Agnode_t * root, int n, ...)
{

  assert (root);

  va_list argp;
  //va_start (argp, root);
  va_start (argp, n);

  char *p;

  Agedge_t *e = first_child (root);
  Agnode_t *result = NULL;

  n--;
  int i = 0;
  for_each_arg (p, char *, argp)
  {
    //printf( "searching for %s\n", p );
    e = resolve_name (e, p);
    if (e)
      {
	result = e->head;
	//printf( "\tfound %s\n", result->name );
	e = first_child (result);
      }
    else
      break;

    if (i == n)
      return result;
    else
      i++;
  }

  va_end (argp);

  //if( i == n )
  // return result;
  //else
  return NULL;
}

inline int
node_typeof (Agnode_t * node, char *type)
{
  char *attr = get_fsattr (node);
  int result = !strcmp (attr, type);
  free (attr);

  return result;
}


Agedge_t *
resolve_name (Agedge_t * h, char *name)
{
  //assert( h );
  if (h == NULL)
    return NULL;

  //printf( "trying to resolve name %s\n", name );

  Agedge_t *e;
  Agraph_t *g = h->head->graph;
  //for_each_edge (e, root, root->graph )
  for (e = h; e; e = agnxtout (g, e))
    {
      Agnode_t *head = e->head;
      //printf( "%s", head->name );
      char *attr = get_fsattr (head);
      //printf( "[%s]\n", attr );
      if (attr != NULL)
	{
	  if (!strcmp (name, attr))
	    {
	      free (attr);
	      return e;
	    }
	  free (attr);
	}
      //__print_ast (g, head, 0);
    }

  return NULL;

}
