#ifndef UTIL_H
#define UTIL_H

#include <gvc.h>
#include <assert.h>
#include <malloc.h>
#define GNU_SOURCE
#include <string.h>
#include <stdarg.h>
#include <linkedlist.h>
#include <parse.h>

typedef struct opts{
  char * outfile;
  keystone_t keystone;
}opts_t;

opts_t parse_opts( int argc, char ** argv );

typedef struct strn
{
  listnode_t node;
  char *str;
} strn_t;

//#define INDENTATION " "
#define INDENTATION "  "
#define indent( fd, depth )                                                                                  \
                  { int i; for( i = 0 ; i < depth ; i++ ) fprintf( fd, INDENTATION ); }                       \

#define printfin( fd, depth, ... ) indent( fd, depth ); fprintf( fd, __VA_ARGS__); fflush( fd )

#define print_asg( g ) __print_ast( g, agfstnode( g ), 0 )
#define print_ast( r ) __print_ast( r->graph, r, 0 )


//#define true  1
//#define false 0


//for each output edge e of root r in graph g
#define for_each_edge( e, r, g )                                                                         \
                  for( e = agfstout(g, r); e ; e = agnxtout(g,e) )                                       \

#define first_child( r )  agfstout( r->graph, r )
#define second_child( r ) agnxtout( r->graph, agfstout( r->graph, r ) )
#define third_child( r )  agnxtout( r->graph, agnxtout( r->graph, agfstout( r->graph, r ) ) )
#define fourth_child( r )  agnxtout( r->graph, agnxtout( r->graph, agnxtout( r->graph, agfstout( r->graph, r ) ) ) )


#define root_graph( r )   r->graph

#define for_each_child( e, r, name )                                                                                      \
               for( e = resolve_name( first_child( r ), #name ) ; e ; e = resolve_name( agnxtout(r->graph,e), #name ) )   \

#define for_each_neighboor( e, g, name )                                                                                  \
               for( e = agnxtout( g, e ); e && node_typeof( e->head, #name) ; e = agnxtout(g,e) )                         \

#define for_child( e, r, name )                                                                                           \
               for( e = resolve_name( first_child( r ), #name ) ; e ; e = NULL )                                          \

#define ROOT( ast )      agfstnode( ast )

#define str_case( a, b )   if ( ! strcmp( a, #b ) )
#define str_elseif( a, b ) else if ( ! strcmp( a, #b ) )
#define str_else           else


inline Agnode_t *get_fschild (Agnode_t * node);
inline Agnode_t *get_scchild (Agnode_t * node);
inline Agnode_t *get_thchild (Agnode_t * node);
inline Agnode_t *get_fhchild (Agnode_t * node);

inline char *get_fsattr (Agnode_t * node);
inline int get_attr (Agnode_t * node, char ***argv);
inline char *get_namedattr (Agnode_t * node, char *name);
static inline int node_type (Agnode_t * root, char *type);
void __print_ast (Agraph_t * g, Agnode_t * root, int depth);
Agnode_t *resolve_path (Agnode_t * root, int n, ...);
Agedge_t *resolve_name (Agedge_t * head, char *name);

//inline void panic( char * error );

#define panic( ... )                                    \
          ({                                             \
          fprintf( stderr, "panic : "   );              \
          fprintf( stderr, __VA_ARGS__  );              \
          fprintf( stderr, " on %s, %s:%d\n", __FILE__, __BASE_FILE__, __LINE__  );              \
          exit(-1);                                     \
          })

#define debug( ... )                                    \
          ({                                             \
          fprintf( stderr, "debug : "   );              \
          fprintf( stderr, __VA_ARGS__  );              \
          fprintf( stderr, " on %s, %s:%d\n", __FILE__, __BASE_FILE__, __LINE__  );              \
          })


#endif
