#include <verilog.h>

static FILE* gfd;

static void
verilog_comments (FILE * fd, char *comments)
{
  fprintf (fd, "// %s", comments);
}


static inline char *
verilog_type (type_t * type)
{

  int size = type_size (type);

  if (size == 1)
    return strdup ("");

  int sign = type_signed (type);

  char *str = NULL;
  asprintf (&str, "%s[%d:0]", sign ? "signed" : "", size - 1);

  return str;

}

static char *
verilog_reserved (module_t * module, char *keyword)
{

  varn_t *node = NULL;
  methodn_t *methodn = NULL;

  for_each_list_element (module->methods, methodn)
  {
    for_each_list_element (methodn->method->variables, node)
    {
      if (!strcmp (keyword, node->var->name))
	{
	  var_t *v = node->var;
	  type_t *type = v->type;
	  char *type_str = verilog_type (type);
	  char *str = NULL;
	  asprintf (&str, "%s_%s", methodn->method->name, v->name);
	  //fprintf (fd, "\treg%s\t%15s;\t//local variable of %s\n", type_str, str, methodn->method->name );
	  return str;

	}
    }
  }

  if (!strcmp (keyword, "wait"))
    return strdup ("wait_0");

  return strdup (keyword);

}


static int
verilog_size (module_t * module, char *var)
{

  varn_t *node;
  methodn_t *methodn;

  for_each_list_element (module->signals, node)
  {
    if (!strcmp (var, node->var->name))
      {
	return type_size (node->var->type);
      }
  }

  for_each_list_element (module->input, node)
  {
    if (!strcmp (var, node->var->name))
      {
	return type_size (node->var->type);
      }
  }

  for_each_list_element (module->output, node)
  {
    if (!strcmp (var, node->var->name))
      {
	return type_size (node->var->type);
      }
  }

  for_each_list_element (module->inout, node)
  {
    if (!strcmp (var, node->var->name))
      {
	return type_size (node->var->type);
      }
  }

  for_each_list_element (module->methods, methodn)
  {
    for_each_list_element (methodn->method->variables, node)
    {
      if (!strcmp (var, node->var->name))
	{
	  return type_size (node->var->type);
	}
    }
  }


  return 0;

}

static void
verilog_trunit_header (FILE * fd, trunit_t * unit)
{
  verilog_comments (fd, "file automaticaly generated by gsc\n\n");
}

static void
verilog_trunit_footer (FILE * fd, trunit_t * unit)
{
}

static inline int
verilog_assigns (Agnode_t * root, char *variable)
{

  if (root == NULL)
    return 0;

  Agedge_t *e;
  int result = 0;
  char *root_str = get_fsattr (root);

  if (!strcmp (root_str, "cBinaryOperator"))
    {
      char *op_str = get_namedattr (root, "op");
      if (!strcmp (op_str, "="))
	{
	  if (verilog_assigns (get_fschild (root), variable))
	    result = 1;
	}
      else if (!strcmp (op_str, "."))
	{
	  if (verilog_assigns (get_fschild (root), variable))
	    {
	      //printf( "hi\n" );
	      char *method_str = get_namedattr (get_scchild (root), "name");
	      if (!strcmp (method_str, "write"))
		result = 1;
	      free (method_str);
	    }
	}
      else
	{
	  result = verilog_assigns (get_fschild (root), variable)
	    || verilog_assigns (get_scchild (root), variable);
	}
      free (op_str);
    }
  else if (!strcmp (root_str, "cNameReference"))
    {
      char *name_str = get_namedattr (root, "name");
      if (!strcmp (name_str, variable))
	result = 1;
    }
  else
    {
      for (e = first_child (root); e; e = agnxtout (root->graph, e))
	{
	  if (verilog_assigns (e->head, variable))
	    {
	      result = 1;
	      break;
	    }
	}
    }

  free (root_str);

  return result;
}

static inline char *
verilog_storage (module_t * module, char *port)
{

  instancen_t *ins;
  for_each_list_element (module->instances, ins)
  {
    portmapn_t *map;
    for_each_list_element (ins->instance->mapping, map)
    {
      if (!strcmp (map->map.wire, port))
	{
	  methodn_t *metn;
	  for_each_list_element (module->methods, metn)
	  {
	    if (verilog_assigns (metn->method->block, map->map.wire))
	      return strdup ("reg");
	  }

	  return strdup ("wire");
	}
    }
  }

  return strdup ("reg");
}

static inline void
verilog_module_header (FILE * fd, module_t * module)
{
  printfin (fd, 0, "module %s( ", module->name);

  varn_t *p;
  char *str = NULL;

  for_each_list_element (module->input, p)
  {
    fprintf (fd, "%s", p->var->name);

    if (LISTNODE (p) != LIST_LAST (module->input))
      fprintf (fd, ", ");
    else if (!LIST_EMPTY (module->inout) || !LIST_EMPTY (module->output))
      fprintf (fd, ", ");


  }

  for_each_list_element (module->inout, p)
  {
    fprintf (fd, "%s", p->var->name);

    if (LISTNODE (p) != LIST_LAST (module->inout))
      fprintf (fd, ", ");
    else if (!LIST_EMPTY (module->output))
      fprintf (fd, ", ");

  }

  for_each_list_element (module->output, p)
  {
    fprintf (fd, "%s", p->var->name);

    if (LISTNODE (p) != LIST_LAST (module->output))
      fprintf (fd, ", ");

  }

  fprintf (fd, " );\n");

  for_each_list_element (module->input, p)
  {
    str = verilog_type (p->var->type);
    printfin (fd, 1, "input  %12s\t%s;\n", str, p->var->name);
    free (str);
  }

  for_each_list_element (module->inout, p)
  {
    str = verilog_type (p->var->type);
    printfin (fd, 1, "inout %10s\t%s;\n", str, p->var->name);
    free (str);
  }

  for_each_list_element (module->output, p)
  {
    str = verilog_type (p->var->type);
    printfin (fd, 1, "output %12s\t%s;\n", str, p->var->name);
    free (str);
  }

  fprintf (fd, "\n");

  for_each_list_element (module->output, p)
  {
    str = verilog_type (p->var->type);

    char *out_str = verilog_storage (module, p->var->name);
    assert (out_str);

    printfin (fd, 1, "%s %12s\t%s;\n", out_str, str, p->var->name);

    free (out_str);
    free (str);
  }

  fprintf (fd, "\n");

}

static inline void
verilog_trunit_declarations (FILE * fd, trunit_t * unit)
{

  enumerationn_t *node;
  for_each_list_element (unit->enumerations, node)
  {
    enumeratorn_t *enumerator;
    int i = 0;
    printfin (fd, 1, "parameter ");
    for_each_list_element (node->enumeration.enumerators, enumerator)
    {
      printfin (fd, 0, "%s = %d", enumerator->enumerator.name, i++);
      if (LISTNODE (enumerator) != LIST_LAST (node->enumeration.enumerators)){
	printfin (fd, 0, ", ");
      }
    }
    printfin (fd, 0, ";\n");
  }

  aliasn_t *alias;
  for_each_list_element (unit->aliases, alias)
  {
    printfin (fd, 1, "//alias %s : %s\n", alias->alias.name,
	     alias->alias.value);
  }

}

static inline void
verilog_declaration(FILE* fd, module_t *module, var_t *var, type_t *type, char* name){
  char *str = verilog_type (type);
  printfin (fd, 1, "reg %s\t %10s", str, name);
  free (str);

  int i;
  for (i = 0; i < var->type->array.n; i++)
    fprintf (fd, "[%d:0]", var->type->array.dim[i] - 1);
  
  fprintf (fd, ";\t");
}

static inline structdef_t*
verilog_var_struct(var_t* var){
  structn_t *structn;
  for_each_list_element(gunit->structs, structn){
    if(!strcmp(structn->structdef.name, var->type->name) ||
       !strcmp(structn->structdef.alias, var->type->name)){
      return &structn->structdef;
    }
  }
  return NULL;    
}


static inline void
verilog_module_declarations (FILE * fd, trunit_t* unit, module_t * module)
{


  //modulen_t *modulen;
  //for_each_list_element (unit->modules, modulen)
  //{
  enumerationn_t *n;
  for_each_list_element (module->enumerations, n)
    {
      enumeratorn_t *enumerator;
      int i = 0;
      printfin (fd, 1, "parameter ");
      for_each_list_element (n->enumeration.enumerators, enumerator)
	{
	  printfin (fd, 0, "%s = %d", enumerator->enumerator.name, i++);
	  if (LISTNODE (enumerator) !=
	      LIST_LAST (n->enumeration.enumerators)){
	    printfin (fd, 0, ", ");
	  }
	}
      printfin (fd, 1, ";\n");
    }
    //}



  varn_t *node = NULL;
  char *name_str = NULL;

  for_each_list_element (module->signals, node)
  {
    var_t *v = node->var;
    type_t *type = v->type;
    char *str = verilog_type (type);

    char *out_str = verilog_storage (module, v->name);

    name_str = verilog_reserved (module, v->name);
    printfin (fd, 1, "%s %s\t%10s", out_str, str, name_str);
    free (name_str);

    int i;
    for (i = 0; i < v->type->array.n; i++)
      fprintf (fd, "[%d:0]", v->type->array.dim[i] - 1);

    fprintf (fd, ";\t//signal\n");

    free (out_str);
    free (str);
  }

  for_each_list_element (module->variables, node)
  {

    structdef_t *structdef = verilog_var_struct(node->var);
    varn_t *varn;

    if(structdef == NULL){
      var_t *v = node->var;
      verilog_declaration(fd, module, node->var, node->var->type, node->var->name);
      fprintf (fd, "\t//variable\n");
    } else {
      for_each_list_element(structdef->fields, varn){
	char * str = NULL;
	asprintf(&str, "%s_%s", node->var->name, varn->var->name);
	verilog_declaration(fd, module, node->var, varn->var->type, str);
	free(str);
	printfin(fd, 1, "//%s struct\n", structdef->name);
      }
    }
    
  }

  methodn_t *methodn = NULL;
  for_each_list_element (module->methods, methodn)
  {
    for_each_list_element (methodn->method->variables, node)
    {
      var_t *v = node->var;
      type_t *type = v->type;
      char *type_str = verilog_type (type);
      char *str = NULL;

      if (!strcmp (v->name, "wait"))
	{
	  asprintf (&str, "%s_%s", methodn->method->name, "wait_0");
	}
      else
	{
	  asprintf (&str, "%s_%s", methodn->method->name, v->name);
	}

      printfin (fd, 1, "reg %s\t%15s", type_str, str);

      int i;
      for (i = 0; i < v->type->array.n; i++)
	fprintf (fd, "[%d:0]", v->type->array.dim[i] - 1);

      fprintf (fd, ";\t//local variable of %s\n", methodn->method->name);

      free (str);
    }
  }

  return;

}

static inline void
verilog_module_footer (FILE * fd, module_t * module)
{
  fprintf (fd, "endmodule\n");
}


static inline void
verilog_method_header (FILE * fd, method_t * method)
{
  assert (method);
  printfin (fd, 1, "always @( ");
  strn_t *node;

  assert (method->sensitivity_def);

  for_each_list_element (method->sensitivity_def, node)
  {
    fprintf (fd, "%s", node->str);

    if (LISTNODE (node) != LIST_LAST (method->sensitivity_def))
      fprintf (fd, " or ");
    if (!LIST_EMPTY (method->sensitivity_pos))
      fprintf (fd, " or ");
  }
  for_each_list_element (method->sensitivity_pos, node)
  {
    // verilator doesn't like posedge(clk)
    // since posedge clk is still valid, there is no harm in changing it
    //fprintf (fd, "posedge(%s)", node->str);
    fprintf (fd, "posedge %s", node->str);

    if (LISTNODE (node) != LIST_LAST (method->sensitivity_pos))
      fprintf (fd, " or ");
    if (!LIST_EMPTY (method->sensitivity_neg))
      fprintf (fd, " or ");
  }
  for_each_list_element (method->sensitivity_neg, node)
  {
    // same thing as in posedge
    //fprintf (fd, "negedge(%s)", node->str);
    fprintf (fd, "negedge %s", node->str);
    if (LISTNODE (node) != LIST_LAST (method->sensitivity_neg))
      fprintf (fd, " or ");

  }

  fprintf (fd, " )\n");
  printfin (fd, 1, "begin : %s\n", method->name);

}

static inline void
verilog_method_footer (FILE * fd, method_t * method)
{
  printfin (fd, 1, "end\n\n");
}


static char *
verilog_assignment (module_t * module, char *variable)
{

  varn_t *v;

  assert (module);
  assert (variable);

  instancen_t *ins;
  for_each_list_element (module->instances, ins)
  {
    portmapn_t *map;
    for_each_list_element (ins->instance->mapping, map)
    {
      if (!strcmp (map->map.wire, variable))
	{
	  return strdup (" = ");
	}
    }
  }

  for_each_list_element (module->output, v)
  {
    assert (v->var);
    assert (v->var->name);

    if (!strcmp (variable, v->var->name))
      {
	return strdup (" <= ");
      }
  }

  for_each_list_element (module->inout, v)
  {
    assert (v->var);
    assert (v->var->name);

    if (!strcmp (variable, v->var->name))
      {
	return strdup (" <= ");
      }
  }

  for_each_list_element (module->signals, v)
  {
    assert (v->var);
    assert (v->var->name);
    if (!strcmp (variable, v->var->name))
      {
	return strdup (" <= ");
      }
  }

  return strdup (" = ");
}

static var_t* verilog_variable(module_t* module, char* name){
  varn_t *varn;
  assert(module);
  assert(name);
  for_each_list_element(module->variables, varn){
    if(!strcmp(varn->var->name, name)){
      return varn->var;
    }
  }
  return NULL;
}

static method_t* verilog_method(module_t* module, char* name){
  methodn_t *methodn;
  assert(module);
  assert(name);
  for_each_list_element(module->methods, methodn){
    if(!strcmp(methodn->method->name, name)){
      return methodn->method;
    }
  }
  return NULL;
}


typedef struct context
{
  int assignment;
  int statement;
  int cbreak;
  int size;
} context_t;

#define context_save( context, stack ) stack = context
#define context_load( context, stack ) context = stack

static void
verilog_code (FILE * fd, module_t * module, method_t * method,
	      Agnode_t * root, int indent)
{

  static context_t context = {.assignment = 0,.statement = 1,.cbreak =
      0,.size = 0
  };
  context_t stack;
  context_save (context, stack);

  if (root == NULL)
    return;

  char *str = get_fsattr (root);

  str_case (str, cBlock)
  {
    Agedge_t *e;
    Agnode_t *last = NULL;
    for_each_edge (e, root, root->graph)
    {

      // handles cCaseLabel special case
      char *fsattr_str = get_fsattr (last);
      if (last != NULL
	  && (!strcmp (fsattr_str, "cCaseLabel")
	      || !strcmp (fsattr_str, "cDefaultLabel")))
	{

	  //free( fsattr_str );

	  if (!context.cbreak)
	    {
	      Agedge_t *elaborate = e;
	      char *next_str = get_fsattr (e->head);
	      Agedge_t *next = (!strcmp ("cCaseLabel", next_str)
				|| !strcmp ("cDefaultLabel",
					    next_str)) ? e : agnxtout (last->
								       graph,
								       e);
	      //Agnode_t * nextn = resolve_path( root, 1, "cCaseLabel");
	      //Agedge_t * next = agfstin( nextn->graph, nextn );

	      Agedge_t *t;
	      for (t = e; t; t = agnxtout (root->graph, t))
		{
		  char *node_str = get_fsattr (t->head);
		  if (!strcmp (node_str, "cCaseLabel")
		      || !strcmp (node_str, "cDefaultLabel"))
		    {
		      next = t;
		      free (node_str);
		      break;
		    }
		  else
		    {
		      free (node_str);
		    }
		}


	      if (next == NULL)
		{
		  free (next_str);
		  if (!context.cbreak)
		    {
		      printfin (fd, indent, "end\n");
		    }
		  //printf( "hey\n" );
		  continue;
		}
	      else
		{
		  //printf( "hi\n" );
		  free (next_str);
		}


	      assert (next);

	      for (elaborate; elaborate;
		   elaborate = agnxtout (last->graph, elaborate))
		{
		  next_str = get_fsattr (elaborate->head);

		  /*
		     if ( strcmp( next_str, "cBreak") ){
		     printfin (fd, indent, "end\n" );
		     e = next;
		     break;
		     }
		   */

		  if (strcmp (next_str, "cCaseLabel")
		      && strcmp (next_str, "cDefaultLabel"))
		    {
		      verilog_code (fd, module, method, elaborate->head,
				    indent + 1);
		      //next = elaborate;
		    }
		  else
		    {
		      verilog_code (fd, module, method,
				    get_fschild (elaborate->head),
				    indent + 1);
		    }

		  free (next_str);

		  if (context.cbreak)
		    {
		      //printf("break!\n");
		      break;
		    }

		}

	      if (context.cbreak)
		{
		  printfin (fd, indent, "end\n");
		}

	      e = next;

	      if (t == NULL || agnxtout (root->graph, t) == NULL)
		{
		  if (!context.cbreak)
		    {
		      printfin (fd, indent, "end\n");
		    }
		  return;
		  //continue;
		}


	    }
	}
      else
	{
	  free (fsattr_str);
	}

      //handles normal cases
      verilog_code (fd, module, method, e->head, indent);
      last = e->head;

      if (agnxtout (root->graph, e) == NULL)
	{
	  fsattr_str = get_fsattr (e->head);
	  assert (fsattr_str);
	  if (!strcmp (fsattr_str, "cCaseLabel")
	      || !strcmp (fsattr_str, "cDefaultLabel"))
	    {
	      if (!context.cbreak)
		{
		  printfin (fd, indent, "end\n");
		}
	    }
	  free (fsattr_str);
	}


    }
  }
  str_elseif (str, cSwitchStatement)
  {
    printfin (fd, indent, "case ( ");
    verilog_code (fd, module, method, get_fschild (root), 0);
    printfin (fd, 0, " )\n");
    verilog_code (fd, module, method, get_scchild (root), indent + 1);
    printfin (fd, indent, "endcase\n");
  }
  str_elseif (str, cForLoop)
  {
    printfin (fd, indent, "for( ");
    context.statement = 0;
    verilog_code (fd, module, method, get_fschild (root), 0);
    printfin (fd, 0, "; ");
    verilog_code (fd, module, method, get_scchild (root), 0);
    printfin (fd, 0, "; ");
    verilog_code (fd, module, method, get_thchild (root), 0);
    context.statement = 1;
    printfin (fd, 0, " )\n");
    printfin (fd, indent, "begin\n");
    verilog_code (fd, module, method, get_fhchild (root), indent + 1);
    printfin (fd, indent, "end\n");
  }
  str_elseif (str, cObjectList)
  {
    Agedge_t *e;
    for_each_child (e, root, cObject)
    {
      verilog_code (fd, module, method, e->head, indent);
    }
  }
  str_elseif (str, cObject)
  {
    if (get_scchild (root))
      {
	var_t *var = var_new (root);
	assert (var);

	char *init_str = get_fsattr (get_scchild (root));
	assert (init_str);

	if (!strcmp (init_str, "cInitializerList"))
	  {
	    int i = 0;
	    Agedge_t *e;
	    for (e = first_child (var->init); e;
		 e = agnxtout (root->graph, e))
	      {
		printfin (fd, indent, "%s_%s[%d] = ", method->name, var->name,
			  i);
		i++;
		verilog_code (fd, module, method, e->head, 0);
		fprintf (fd, ";\n");
	      }
	  }
	else
	  {
	    printfin (fd, indent, "%s_%s = ", method->name, var->name);
	    verilog_code (fd, module, method, var->init, 0);
	    fprintf (fd, ";\n");
	  }
	free (init_str);
	var_free (var);
      }
  }
  str_elseif (str, cCaseLabel)
  {
    verilog_code (fd, module, method, get_scchild (root), indent);
    char *str = get_fsattr (get_fschild (root));
    context.cbreak = 0;
    if (!strcmp (str, "cCaseLabel"))
      {
	printfin (fd, 0, ", ");
	verilog_code (fd, module, method, get_fschild (root), 0);
      }
    else
      {
	printfin (fd, 0, " : begin\n");
	verilog_code (fd, module, method, get_fschild (root), indent + 1);
      }

    free (str);

    if (context.cbreak)
      {
	printfin (fd, indent, "end\n", indent);
      }

  }
  str_elseif (str, cDefaultLabel)
  {
    char *str = get_fsattr (get_fschild (root));
    context.cbreak = 0;

    printfin (fd, indent, "default : begin\n");
    verilog_code (fd, module, method, get_fschild (root), indent + 1);
    free (str);

    if (context.cbreak)
      {
	printfin (fd, indent, "end\n", indent);
      }
  }
  str_elseif (str, cBreak)
  {
    context.cbreak = 1;
  }
  str_elseif (str, cReturn)
  {
    //makes sure we are inside a normal method, not a systemc process
    assert (!method->sc_method);
    context.cbreak = 1;
    printfin (fd, indent, "%s = ", method->name);
    verilog_code (fd, module, method, get_fschild (root), 0);
    printfin (fd, 0, ";\n");
    //context.cbreak = 1;
  }
  str_elseif (str, cIfStatement)
  {
    printfin (fd, indent, "if (");
    verilog_code (fd, module, method, get_fschild (root), 0);
    printfin (fd, 0, ") begin\n");
    verilog_code (fd, module, method, get_scchild (root), indent + 1);
    Agnode_t *third = get_thchild (root);
    if (third)
      {
	printfin (fd, indent, "end\n");
	printfin (fd, indent, "else begin\n");
	verilog_code (fd, module, method, third, indent + 1);
      }
    printfin (fd, indent, "end\n");
  }
  str_elseif (str, cUnaryOperator)
  {
    char *op = get_namedattr (root, "op");
    assert (op);

    str_case (op, ())
    {
      Agnode_t *cast;

      cast =
	resolve_path (root, 4, "cInstantiatedForwardType",
		      "cTemplateArgumentList", "cLiteral", "cBuiltInType");
      if (!cast)
	cast = resolve_path (root, 1, "cBuiltInType");

      if (!cast)
	cast = resolve_path (root, 1, "cForwardType");

      if (cast)
	{
	  //printfin( fd , 0 , " CAST ");
	}
      else
	{
	  printfin (fd, 0, "( ");
	  //print_ast( root );
	  verilog_code (fd, module, method, get_fschild (root), 0);
	  printfin (fd, 0, " )");
	}

    }
    str_elseif (op, -)
    {
      printfin (fd, 0, "-");
      verilog_code (fd, module, method, get_fschild (root), 0);
    }
    str_elseif (op, !)
    {
      printfin (fd, 0, "!");
      verilog_code (fd, module, method, get_fschild (root), 0);
    }
    str_elseif (op, ~)
    {
      printfin (fd, 0, "~");
      verilog_code (fd, module, method, get_fschild (root), 0);
    }
    str_elseif (op, ++)
    {

      Agnode_t *child = get_fschild (root);
      char *var_str = NULL;
      var_str = get_namedattr (child, "name");

      verilog_code (fd, module, method, get_fschild (root), indent);

      char *assign_str = verilog_assignment (module, "teste");

      //if (assign_str)
      //{
      printfin (fd, 0, " %s ", assign_str);
      free (assign_str);
      //}
      //else
      //{
      //printfin (fd, 0, " <= ");
      //}

      verilog_code (fd, module, method, get_fschild (root), 0);
      printfin (fd, 0, " + 1");
      //verilog_code (fd, module, method, get_fschild (root), 0);

      if (context.statement)
	{
	  printfin (fd, 0, ";\n");
	}

      free (var_str);
    }
    str_elseif (op, --)
    {

      Agnode_t *child = get_fschild (root);
      char *var_str = NULL;
      var_str = get_namedattr (child, "name");

      verilog_code (fd, module, method, get_fschild (root), indent);

      char *assign_str = verilog_assignment (module, "teste");

      printfin (fd, 0, " %s ", assign_str);
      free (assign_str);
      verilog_code (fd, module, method, get_fschild (root), 0);
      printfin (fd, 0, " - 1");
      
      if (context.statement)
	{
	  printfin (fd, 0, ";\n");
	}

      free (var_str);
    }
    str_else
    {
      print_ast (root);
      panic ("unknow unary operator %s !\n", op);
    }

    free (op);
  }
  str_elseif (str, cBinaryOperator)
  {


    char *op = get_namedattr (root, "op");
    assert (op);

    str_case (op, ())
    {
      int writeop = 0;
      char *node_str = get_fsattr (get_fschild (root));
      assert (node_str);

      if (!strcmp (node_str, "cBinaryOperator"))
	{
	  Agnode_t *fschild = get_fschild (root);
	  char *op_str = get_namedattr (fschild, "op");
	  assert (op_str);
	  if (!strcmp (op_str, "."))
	    {
	      char *name_str = get_namedattr (get_scchild (fschild), "name");
	      assert (name_str);
	      if (context.statement)
		if (!strcmp (name_str, "write"))
		  writeop = 1;
	      free (name_str);
	    }
	  free (op_str);
	}

      context.assignment = writeop;
      verilog_code (fd, module, method, get_fschild (root), indent);
      context.assignment = stack.assignment;

      //if (writeop)
      //fprintf( fd, " **%d** ", context.size  );

      Agnode_t *methodn;

      methodn = resolve_path (root, 1, "cBinaryOperator");

      if (!methodn)
	{
	  //print_ast( root );
	  methodn = resolve_path (root, 1, "cNameReference");
	}
      
      char * str = get_namedattr(get_fschild(root), "name");
      if(str != NULL){
	method_t * method = verilog_method(module, str);
	free(str);
      
	if(method != NULL){
	  int size = type_size (method->returns);
	  if (size == 0) 
	    {
	      printfin(fd, 0, "(");
	      Agnode_t *args = get_scchild(root);
	      if (args)
		{
		  Agedge_t *e;
		  for_each_edge (e, args, args->graph)
		    {
		      printfin (fd, 0, "%s", e == first_child (args) ? "" : ", ");
		      verilog_code (fd, module, method, e->head, 0);
		    }
		}
	      printfin (fd, 0, ");\n");
	      return;
	    }	  
	}
      }

      assert (methodn);

      char *op_f = "( ";
      char *cl_f = " )";
      char *ar_s = ", ";

      Agnode_t *scchild = get_scchild (methodn);


      if (scchild)
	{
	  assert (scchild);

	  char *scchild_str = get_fsattr (scchild);
	  assert (scchild_str);
	  if (!strcmp (scchild_str, "cNameReference"))
	    {
	      char *name_str = get_namedattr (scchild, "name");
	      /*
	         According to
	         http://www.asic-world.com/verilog/verilog2k2.html

	         verilog 2001 does support multi dimensional arrays, while
	         verilog 1995 does not.

	         Please, make sure you are using a 2001 compatible verilog
	         compiler.
	       */
	      if (!strcmp (name_str, "range"))
		{
		  op_f = "[";
		  cl_f = "]";
		  ar_s = ":";
		}
	      free (name_str);
	    }
	  free (scchild_str);
	}

      Agnode_t *args = resolve_path (root, 1, "cFunctionArgumentList");
      if (args)
	{
	  printfin (fd, 0, "%s", op_f);
	  Agedge_t *e;
	  for_each_edge (e, args, args->graph)
	  {
	    printfin (fd, 0, "%s", e == first_child (args) ? "" : ar_s);
	    verilog_code (fd, module, method, e->head, 0);
	  }
	  printfin (fd, 0, "%s", cl_f);
	}

      if (writeop)
	fprintf (fd, ";\n");

    }
    str_else
    {

      str_case (op, =) context.assignment = true;
      str_else if (!context.assignment)
	  context.assignment = false;

      verilog_code (fd, module, method, get_fschild (root), indent);
      context.assignment = stack.assignment;

      str_case (op,[])
      {
	//print_ast( root );
	//verilog_code( fd, module, method, get_fschild( root ), 0 );
	printfin (fd, 0, "[");
	context_save (context, stack);
	verilog_code (fd, module, method, get_scchild (root), 0);
	context_load (context, stack);
	printfin (fd, 0, "]");
      }
      str_elseif (op, (cast))
      {
	verilog_code (fd, module, method, get_scchild (root), 0);
      }
      str_elseif (op, =)
      {

	Agnode_t *child = get_fschild (root);
	char *var_str = NULL;

	var_str = get_namedattr (child, "name");

	if (var_str == NULL)
	  {
	    child =
	      resolve_path (root, 2, "cBinaryOperator", "cNameReference");
	    var_str = get_namedattr (child, "name");
	  }

	if (var_str == NULL)
	  {
	    child =
	      resolve_path (root, 3, "cBinaryOperator", "cBinaryOperator",
			    "cNameReference");
	    var_str = get_namedattr (child, "name");
	  }

	if (var_str == NULL)
	  {
	    panic ("can't understand this assingment pattern!");
	  }


	assert (var_str);
	char *assignment = verilog_assignment (module, var_str);
	free (var_str);

	printfin (fd, 0, " %s ", assignment);
	free (assignment);

	//fprintf( fd, " **%d** ", context.size  );

	verilog_code (fd, module, method, get_scchild (root), 0);
	if (context.statement)
	  {
	    printfin (fd, 0, ";\n");
	  }

      }
      str_elseif (op,.)
      {

	Agnode_t *schild = get_scchild (root);
	assert (schild);
	char *nameref = get_namedattr (schild, "name");
	assert (op);

	char *name = get_namedattr(get_fschild(root), "name");
	if(name != NULL){
	  var_t * var = verilog_variable(module, name);
	  if(var != NULL){
	    structdef_t * def = verilog_var_struct(var);
	    char *field = get_namedattr(get_scchild(root), "name");
	    if(def != NULL){
	      printfin (fd, 0, "_%s", field);
	      free(name);
	      free(field);
	      return;
	    }
	    free(field);
	  }
	  
	}
	free(name);
	

	str_case (nameref, write)
	{
	  // interpret write as an assigment
	  Agnode_t *child = get_fschild (root);
	  char *var_str = NULL;

	  var_str = get_namedattr (child, "name");

	  if (var_str == NULL)
	    {
	      child =
		resolve_path (root, 2, "cBinaryOperator", "cNameReference");
	      var_str = get_namedattr (child, "name");
	    }

	  char *assignment = verilog_assignment (module, var_str);
	  free (var_str);

	  //if (assignment)
	  //{
	  printfin (fd, 0, " %s ", assignment);
	  //}
	  //else
	  //{
	  //  printfin (fd, 0, " = ");
	  //}

	  free (assignment);
	  //printfin( fd , 0 , " = " );
	}
	str_elseif (nameref, event)
	{
	  // interpret write as an assigment
	  //Agnode_t *fschild = get_fschild( root );
	  //char *child_str = get_namedattr( fschild, "name" );
	  //printfin( fd , 0 , " == %s", child_str );
	  printfin (fd, 0, " == ");
	  verilog_code (fd, module, method, get_fschild (root), 0);
	  //free( child_str );
	}
	str_elseif (nameref, read)
	{
	  // do nothing
	}
	str_elseif (nameref, range)
	{
	  //printfin( fd, 0, ".range");
	  // do nothing
	}
	str_else
	{
	  printfin (fd, 0, " %s ", op);
	  verilog_code (fd, module, method, get_scchild (root), 0);
	}
	free (nameref);
      }
      str_else
      {
	printfin (fd, 0, " %s ", op);
	verilog_code (fd, module, method, get_scchild (root), 0);
      }
    }

    free (op);
  }
  str_elseif (str, cTernaryOperator)
  {

    verilog_code (fd, module, method, get_fschild (root), 0);
    printfin (fd, 0, " ? ");
    verilog_code (fd, module, method, get_scchild (root), 0);
    printfin (fd, 0, " : ");
    verilog_code (fd, module, method, get_thchild (root), 0);
  }
  str_elseif (str, cNameReference)
  {

    char *name = get_namedattr (root, "name");

    char *keyword = verilog_reserved (module, name);

    context.size = verilog_size (module, name);

    //if (keyword)
    //{
    printfin (fd, indent, "%s", keyword);
    free (keyword);
    //  }
    //else
    //{
    //printfin (fd, indent, "%s", name);
    //}

    //if ( context.assignment ) { printfin( fd, 0, "ASSIGNMENT"); }

    free (name);
  }
  str_elseif (str, cLiteral)
  {
    char *name = get_namedattr (root, "value");

    if (name[0] == '0' && name[1] == 'x')
      {
	//this is an hexa number
	printfin (fd, indent, "'h%s", &name[2]);
      }
    else if (name[strlen (name) - 1] == 'L')
      {
	// this is a long number
	char *s = (char *) strndup (name, strlen (name) - 1);
	s[strlen (name) - 1] = '\0';
	printfin (fd, indent, "%s", s);
	free (s);
      }
    else
      {
	printfin (fd, indent, "%s", name);
      }
    free (name);

  }
  str_else
  {
    print_ast (root);
    panic ("can't translate %s to verilog ", str);
  }

  free (str);

  return;
}


static inline void
verilog_function_header (FILE * fd, method_t * method)
{
  assert (method);

  char *str;

  // !!! FOR SOME REASON, ICARUS VERILOG DOESN'T ACCEPT SIGNED RETURN VALUES FOR FUNCTIONS !!!//
  //str = verilog_type( method->returns );
  int size = type_size (method->returns);
  if (size == 0) {
    printfin (fd, 1, "task %s;\n", method->name);
  }
  else if (size == 1) {
    printfin (fd, 1, "function %s;\n", method->name);
  }
  else {
    printfin (fd, 1, "function [%d:0] %s;\n", size - 1, method->name);
  }
  //free( str );

  varn_t *node;
  for_each_list_element (method->parameters, node)
  {
    str = verilog_type (node->var->type);
    printfin (fd, 2, "input %s %s;\n", str, node->var->name);
    free (str);
  }

  printfin (fd, 1, "begin\n");
}

static void
verilog_function_body (FILE * fd, module_t * module, method_t * method)
{
  verilog_code (fd, module, method, method->block, 2);
}

static inline void
verilog_function_footer (FILE * fd, method_t * method)
{
  printfin (fd, 1, "end\n");
  int size = type_size (method->returns);
  if (size == 0) {
    printfin (fd, 1, "endtask\n\n");
  }
  else{
    printfin (fd, 1, "endfunction\n\n");
  }
}


static void
verilog_method_body (FILE * fd, module_t * module, method_t * method)
{

  verilog_code (fd, module, method, method->block, 2);

}

static inline void
verilog_instance (FILE * fd, instance_t * instance)
{
  printfin (fd, 1, "%s %s(\n", instance->module, instance->name);

  portmapn_t *node;
  for_each_list_element (instance->mapping, node)
  {
    portmap_t map = node->map;
    printfin (fd, 2, ".%s(%s)", map.port, map.wire);

    if (LISTNODE (node) != LIST_LAST (instance->mapping)){
      printfin (fd, 0, ",");
    }

    printfin (fd, 0, "\n");
  }

  printfin (fd, 1, ");\n\n");

}


void
verilog (FILE * fd, trunit_t * unit)
{

  //gunit = unit;
  gfd = fd;


  verilog_trunit_header (fd, unit);

  modulen_t *n;
  for_each_list_element (unit->modules, n)
  {
    module_t *module = n->module;

    if (!module_complete (module))
      panic("incomplete module implementation : are you sure you've written all declared methods?");

    verilog_module_header (fd, module);


    verilog_trunit_declarations (fd, unit);

    verilog_module_declarations (fd, unit, module);

    methodn_t *m;
    for_each_list_element (module->methods, m)
    {
      method_t *method = m->method;

      fprintf (fd, "\n");
      if (!method->sc_method)
	{
	  verilog_function_header (fd, method);
	  verilog_function_body (fd, module, method);
	  verilog_function_footer (fd, method);
	}
      else
	{
	  verilog_method_header (fd, method);
	  verilog_method_body (fd, module, method);
	  verilog_method_footer (fd, method);
	}
    }

    //verilog_method_body (fd, module, module->ctor);
    //verilog_method_header (fd, module->ctor);
    //printf("\tctor(){\n");
    //verilog_method_footer (fd, module->ctor);
    //printf("\t}\n");

    instancen_t *ins;
    for_each_list_element (module->instances, ins)
    {
      verilog_instance (fd, ins->instance);
    }

    verilog_module_footer (fd, module);

    fprintf (fd, "\n");
  }

  verilog_trunit_footer (fd, unit);

  return;
}
