#include <vhdl.h>

static void
vhdl_comments (FILE * fd, char *comments)
{
  fprintf (fd, "-- %s", comments);
}


static inline char *
vhdl_type (type_t * type)
{
  if (type->name)
    {
      if (!strcmp (type->name, "int"))
	return strdup ("integer");
      if (!strcmp (type->name, "bool"))
	return strdup ("std_logic");
      if (!strcmp (type->name, "sc_uint"))
	return strdup ("integer");
      if (!strcmp (type->name, "sc_int"))
	return strdup ("integer");
      else
	{
	  //panic( "can't translate %s to vhdl\n", type->name );
	  return strdup (type->name);
	}
    }
  else
    {
      print_ast (type->implementation);
      panic ("unknow type name %s", type->name);
      return;
    }
  return NULL;
}

static void
vhdl_trunit_header (FILE * fd, trunit_t * unit)
{
  //fprintf (fd, "-- file automaticaly generated by gsc\n\n");
  vhdl_comments (fd, "file automaticaly generated by gsc\n\n");
  fprintf (fd, "library std;\n");
  fprintf (fd, "use std.textio.all;\n");
  fprintf (fd, "library ieee;\n");
  fprintf (fd, "use ieee.std_logic_1164.all;\n");
  fprintf (fd, "use ieee.std_logic_unsigned.all;\n");
  fprintf (fd, "use ieee.std_logic_arith.all;\n");
  fprintf (fd, "use ieee.std_logic_signed.all;\n");
  fprintf (fd, "\n\n");

}

static void
vhdl_trunit_footer (FILE * fd, trunit_t * unit)
{
}

static inline void
vhdl_module_header (FILE * fd, module_t * module)
{
  //module header
  fprintf (fd, "entity %s is\n\n", module->name);
  fprintf (fd, "\tport(\n");


  varn_t *p;
  char *str = NULL;

  for_each_list_element (module->input, p)
  {
    str = vhdl_type (p->var->type);
    fprintf (fd, "\t\t%s\t: in\t%s", p->var->name, str);
    free (str);

    if (LISTNODE (p) != LIST_LAST (module->input))
      fprintf (fd, ";\n");
    else if (!LIST_EMPTY (module->output))
      fprintf (fd, ";\n");


  }

  for_each_list_element (module->output, p)
  {
    str = vhdl_type (p->var->type);
    fprintf (fd, "\t\t%s\t: out\t%s", p->var->name, str);
    free (str);

    if (LISTNODE (p) != LIST_LAST (module->output))
      fprintf (fd, ";\n");

  }

  fprintf (fd, "\n\t);\n\n");
  fprintf (fd, "end %s;\n\n", module->name);
  fprintf (fd, "\n");

}

static inline void
vhdl_trunit_declarations (FILE * fd, trunit_t * unit)
{

  enumerationn_t *node;
  for_each_list_element (unit->enumerations, node)
  {
    enumeratorn_t *enumerator;
    printf ("\tenumeration %s : { ", node->enumeration.name);
    for_each_list_element (node->enumeration.enumerators, enumerator)
    {
      printf ("%s ", enumerator->enumerator.name);
    }
    printf ("}\n");
  }

  aliasn_t *alias;
  for_each_list_element (unit->aliases, alias)
  {
    printf ("\talias %s : %s\n", alias->alias.name, alias->alias.value);
  }

}

static inline void
vhdl_module_declarations (FILE * fd, module_t * module)
{

  varn_t *node = NULL;

  for_each_list_element (module->signals, node)
  {
    var_t *v = node->var;
    type_t *type = v->type;
    char *str = vhdl_type (type);
    fprintf (fd, "\tsignal %15s : %s;\n", v->name, str);
    free (str);
  }

  for_each_list_element (module->variables, node)
  {
    var_t *v = node->var;
    type_t *type = v->type;
    char *str = vhdl_type (type);
    fprintf (fd, "\tvariable %15s : %s;\n", v->name, str);
    free (str);
  }

  return;

}

static inline void
vhdl_module_footer (FILE * fd, module_t * module)
{
  //fprintf (fd, "endmodule\n");
}


static inline void
vhdl_method_header (FILE * fd, method_t * method)
{
  assert (method);
  fprintf (fd, "\t%s:\n", method->name);
  fprintf (fd, "\tprocess( ");
  //fprintf (fd, " sensitivity list ");
  //char *node;
  strn_t *node;

  //if ( ! LIST_EMPTY( method->sensitivity_def) ){
  assert (method->sensitivity_def);

  for_each_list_element (method->sensitivity_def, node)
  {
    printf ("%s", node->str);
    if (LISTNODE (node) != LIST_LAST (method->sensitivity_def))
      fprintf (fd, ", ");

  }
  //for_each_list_element (method->sensitivity_pos, node)
  //{
  // printf ("posedge(%s) or ", node->str);
  //}
  //for_each_list_element (method->sensitivity_neg, node)
  //{
  //  printf ("negedge(%s) or ", node->str);
  //}

  fprintf (fd, " )\n");
  fprintf (fd, "\tbegin\n");
}

static inline void
vhdl_method_footer (FILE * fd, method_t * method)
{
  fprintf (fd, "\tend process;\n\n");
}


static void
vhdl_code (FILE * fd, module_t * module, Agnode_t * root, int indent)
{
  //printf( "hi\n" );
  //print_ast( node );

  if (root == NULL)
    return;

  char *str = get_fsattr (root);

  str_case (str, cBlock)
  {
    //vhdl_code( fd, module, get_fschild( root ), indent + 1 );
    Agedge_t *e;
    for_each_edge (e, root, root->graph)
    {
      //print_ast( e->head );
      vhdl_code (fd, module, e->head, indent + 1);
      //printfin( fd, 0, ";\n");
    }
  }
  str_elseif (str, cSwitchStatement)
  {
    printfin (fd, indent, "case ");
    vhdl_code (fd, module, get_fschild (root), 0);
    printfin (fd, 0, " is\n");
    vhdl_code (fd, module, get_scchild (root), indent);
    printfin (fd, indent, "end case;\n");
  }
  str_elseif (str, cCaseLabel)
  {
    printfin (fd, indent, "when ");
    vhdl_code (fd, module, get_scchild (root), 0);
    printfin (fd, 0, " =>\n");
    vhdl_code (fd, module, get_fschild (root), indent);
  }
  str_elseif (str, cDefaultLabel)
  {
    printfin (fd, indent, "default :\n");
  }
  str_elseif (str, cBreak)
  {
    printfin (fd, indent, "break;\n");
  }
  str_elseif (str, cFunctionArgumentList)
  {
    Agedge_t *e = NULL;
    for_each_edge (e, root, root->graph)
    {

      printf ("%s", e == first_child (root) ? " " : ", ");

      vhdl_code (fd, module, e->head, indent);
    }
    //vhdl_code( fd, module, get_fschild( root ), indent );
  }
  str_elseif (str, cIfStatement)
  {
    printfin (fd, indent, "if (");
    vhdl_code (fd, module, get_fschild (root), 0);
    printfin (fd, 0, ") then\n");
    vhdl_code (fd, module, get_scchild (root), indent);
    Agnode_t *third = get_thchild (root);
    if (third)
      {
	printfin (fd, indent, "else\n");
	vhdl_code (fd, module, third, indent);
      }
    printfin (fd, indent, "end if;\n");
  }
  str_elseif (str, cBinaryOperator)
  {

    //print_ast( root );
    vhdl_code (fd, module, get_fschild (root), indent);
    char *op = get_namedattr (root, "op");
    assert (op);

    str_case (op, ())
    {
      Agnode_t *args = resolve_path (root, 1, "cFunctionArgumentList");
      if (args)
	{
	  printfin (fd, 0, "(");
	  //print_ast( args );
	  vhdl_code (fd, module, args, 0);
	  printfin (fd, 0, ")");
	}
    }
    str_elseif (op,[])
    {
      //print_ast( root );
      //vhdl_code( fd, module, get_fschild( root ), 0 );
      printfin (fd, 0, "[");
      vhdl_code (fd, module, get_scchild (root), 0);
      printfin (fd, 0, "]");
    }
    str_elseif (op, =)
    {
      printfin (fd, 0, " = ");
      vhdl_code (fd, module, get_scchild (root), 0);
      printfin (fd, 0, ";\n");
    }
    str_else
    {
      printfin (fd, 0, "%s", op);
      vhdl_code (fd, module, get_scchild (root), 0);
    }

    //if( !strcmp( "=" , op ) )
    //printf( "" );
    free (op);
  }
  str_elseif (str, cNameReference)
  {
    char *name = get_namedattr (root, "name");
    printfin (fd, indent, "%s", name);
    free (name);
  }
  str_elseif (str, cLiteral)
  {
    char *name = get_namedattr (root, "value");
    printfin (fd, indent, "%s", name);
    free (name);
  }
  str_else
  {
    panic ("can't translate %s to vhdl ", str);
  }

  free (str);

  return;
}


static void
vhdl_method_body (FILE * fd, module_t * module, method_t * method)
{
  fprintf (fd, "\t\t--method implementation\n");
  vhdl_code (fd, module, method->block, 1);
  //print_ast( method->implementation );
}

static inline void
vhdl_instance (FILE * fd, instance_t * instance)
{
  fprintf (fd, "\tinstance %s of %s{\n", instance->name, instance->module);

  portmapn_t *node;
  for_each_list_element (instance->mapping, node)
  {
    portmap_t map = node->map;
    fprintf (fd, "\t\t%s => %s\n", map.port, map.wire);
  }

  fprintf (fd, "\t}\n");

}


void
vhdl (FILE * fd, trunit_t * unit)
{

  vhdl_trunit_header (fd, unit);

  modulen_t *n;
  for_each_list_element (unit->modules, n)
  {
    module_t *module = n->module;

    vhdl_module_header (fd, module);

    fprintf (fd, "architecture rtl of %s is\n", module->name);

    vhdl_trunit_declarations (fd, unit);

    vhdl_module_declarations (fd, module);

    fprintf (fd, "begin\n\n");
    methodn_t *m;
    for_each_list_element (module->methods, m)
    {
      method_t *method = m->method;

      if (!method->sc_method)
	continue;
      vhdl_method_header (fd, method);

      vhdl_method_body (fd, module, method);

      vhdl_method_footer (fd, method);
    }

    //vhdl_method_body (fd, module, module->ctor);
    //vhdl_method_header (fd, module->ctor);
    //printf("\tctor(){\n");
    //vhdl_method_footer (fd, module->ctor);
    //printf("\t}\n");

    instancen_t *ins;
    for_each_list_element (module->instances, ins)
    {
      vhdl_instance (fd, ins->instance);
    }

    fprintf (fd, "end rtl;\n");

    vhdl_module_footer (fd, module);
  }

  vhdl_trunit_footer (fd, unit);

  return;
}
