// Generated by CynthVLG 2.5.1  (091704)
// Fri Mar 24 23:00:40 2006
// from avalon_mp3.cpp
      
module avalon_mp3(clk, reset_n, address, read, readdata, write, writedata, waitrequest, chipselect);

      input clk;
      input reset_n;
      input [2:0] address;
      input read;
      input write;
      input [31:0] writedata;
      input chipselect;
      output [31:0] readdata;
      reg [31:0] readdata;
      output waitrequest;
      reg waitrequest;
      reg fifo_write_block;
  // INPUT FROM AVALON BUS
/*
  // SAMPLE OUT TO DAC
  sc_out<sc_uint<1> >   MAddr;
  sc_out<sc_uint<3> >   MCmd;
  sc_out<sc_uint<32> >  MData;
  sc_in<bool>          SCmdAccept;
  sc_in<sc_uint<8> >   SData;
  sc_in<sc_uint<2> >   SResp;
*/
      reg[31:0] fifo_block[127:0];
      reg[31:0] fifo_data[127:0];
      reg[6:0] fifo_begin;
      reg[6:0] fifo_end;

/*
= DOCUMENTACAO: =

(OBS: para entender o barramento avalon, leia:
https://atlas.lsc.ic.unicamp.br/ApoloMp3/attachment/wiki/Repositorio/mnl_avalon_bus.pdf
)

Existem 2 fifos, uma para dados e uma para blocos.
As duas fifos sao sincronizadas e por isso existe apenas 1 controle
para as duas fifos.
Dessa maneira, as duas fifos tem tamanhos iguais, elas crescem e so
consumidas de forma igual. Quando falta n elementos para encher uma
fifo, tambem faltam n elementos para encher a outra fifo.

O controle das fifos  feito atraves dos ponteiros (enderecos):
fifo_begin -> comeco dos dados
fifo_end -> fim dos dados

De maneira que:


considere # = dados
          . = espao vazio



  |...................#######################...............|
  ^                   ^                      ^              ^
  zero                fifo_begin             fifo_end       FIFOSIZE
 
  Se( fifo_end > fifo_begin )
    Espaco na fifo = FIFOSIZE - fifo_end + fifo_begin - 1;



  |##################...............########################|
  ^                  ^              ^                       ^
  zero               fifo_end       fifo_begin              FIFOSIZE
 
  Se( fifo_end < fifo_begin )
    Espaco na fifo = fifo_begin - fifo_end - 1;





  |##########################.##############################|
  ^                          ^^                             ^
  zero               fifo_end fifo_begin                    FIFOSIZE


  |########################################################.|
  ^                                                       ^
  fifo_begin                                              fifo_end
 
  Se( (fifo_end+1)%FIFOSIZE == fifo_begin )
    A fifo est cheia


  Se( fifo_end == fifo_begin )
    A fifo est vazia



Assim,sempre vai sobrar um slot vazio na fifo, para se poder
diferenciar o caso em que a fifo esta vazia do caso em que a fifo 
esta cheia.


Uma escrita na fifo  feita da seguinte maneira:
fifo[fifo_end] = dados;
fifo_end = (fifo_end + 1) % fifo_size;




=== O barramento fazendo uma leitura no avalon_mp3: ===

O registrador zero indica se a fifo tem algum espao sobrando e o 
software pode enviar mais um dado.
Portanto, seguindo o esquema do barramento avalon:
  Se( chipselect )
    Se( read )
      Caso address == 0 {
        read_data = "Existe algum espao na fifo?(sim ou nao)"
      }
Como a sada read_data  registrada (sensivel ao clock), 
a resposta s estar pronta para ser lida pelo barramento na
prxima subida de clock.

Logo, ser preciso um wait_state do barramento no caso da leitura.
Veja na pgina 29 e 30 do documento:
https://atlas.lsc.ic.unicamp.br/ApoloMp3/attachment/wiki/Repositorio/mnl_avalon_bus.pdf

       _   _   _   _   _ 
clk  _| |_| |_| |_| |_| |_
          ^   ^  
          1   2 

1 => Pedido de leitura pelo barramento (chip_select = 1, read = 1)
2 => Dados vlidos na sada (read_data)

*/
//#define DEBUGFLAG 0
//#define dprintf(...) if (DEBUGFLAG) fprintf(stdout, __VA_ARGS__)
//#define dprintvar(var) if ( DEBUGFLAG ) { cout << var << "\n"; }
         always @(posedge(clk))
          begin :prc_avalon
            if (!reset_n) begin
               fifo_write_block <= 1'd0;
               fifo_begin <= 7'd0;
               fifo_end <= 7'd0;
            end
            else begin
               waitrequest <= 1'd0;
               if (chipselect) begin
                  if (read) begin
                     case (address) 

                        3'd0: begin // Buffer full?
                           readdata <= {31'b0, ({1'b0, fifo_end} + 8'd1 & 8'd127) != {1'b0, fifo_begin}};
                        end
                        
                        3'd1: begin // Buffer empty?
                           readdata <= {31'b0, !(fifo_end == fifo_begin)};
                        end
                        
                        3'd2: begin // Read data from fifo
                           readdata <= fifo_data[fifo_begin];
                        end
                        
                        3'd3: begin // Read block from fifo
                           readdata <= fifo_block[fifo_begin];
                           fifo_begin <= fifo_begin + 7'd1 & 7'd127;
                        end
                        
                        3'd4: begin // Manual Reset
                           fifo_write_block <= 1'd0;
                           fifo_begin <= 7'd0;
                           fifo_end <= 7'd0;
                        end
                        
                        3'd5: begin // Read fifo_begin value (debug)
                           readdata <= {25'b0, fifo_begin};
                        end
                        
                        3'd6: begin // Read fifo_end value (debug)
                           readdata <= {25'b0, fifo_end};
                        end
                        
                     endcase

                  end
                  if (write) begin
                     fifo_write_block <= !fifo_write_block;
                     if (!fifo_write_block) begin // Data write
                        fifo_data[fifo_end] = writedata;
                     end
                     else begin
                        fifo_block[fifo_end] = writedata;
                        fifo_end <= fifo_end + 7'd1 & 7'd127;
                     end
                  end
               end
            end
         end


endmodule

