#ifndef IMDCTWINDOW_H
#define IMDCTWINDOW_H
#include <systemc.h>
#include "multiplier.h"
#include "imdctwindow_buf.h"

// Estados
SC_MODULE(imdctwindow)
{
enum state_t {read_state, read_state_w, write_state_www, write_state_wwww, buf_mult, buf_mult_addr, buf_mult_w, buf_mult_ww, write_state, write_ram, bt2_state };

  sc_in<bool>          clk;
  sc_in<bool>          Reset_n;

  // Sinais do IMDCT
  sc_in<sc_uint<32> >  data_in;
  sc_in<sc_uint<2> >   in_nch; 	
  sc_in<sc_uint<2> >   in_block_type;
  sc_in<bool>          in_trans;
  sc_out<bool>         in_ready;
    
  // Sinais do Overlap
  sc_out<sc_uint<32> > data_out;
  sc_out<sc_uint<2> >  out_nch;
  sc_out<bool>         out_trans;
  sc_in<bool>          out_ready;
  
  // Sinais internos
  sc_signal<sc_uint<6> >  i;
  sc_signal<sc_uint<16> >  j;  
  sc_signal<sc_uint<2> >  nch;
  sc_signal<sc_uint<2> >  block_type;
  sc_signal<state_t>      state;
  sc_signal<state_t>      next_state;    
  sc_signal<sc_uint<32> > data_ant;
  sc_signal<bool >        read_on_ram;
  sc_signal<sc_uint<32> > mul_c_ant;
  sc_signal<sc_uint<32> > mul_cc;
  sc_signal<bool >        prox_read;
  
  sc_signal<bool >        wait;
  sc_signal<bool >        bt2;

  // Buffer -> memoria ROM
  sc_signal<sc_uint<32> > buf_rom_out;
  sc_signal<sc_uint<6> >  buf_rom_addr;
  
  // Buffer -> memoria RAM
  sc_signal<sc_uint<32> > buf_ram_in;
  sc_signal<sc_uint<32> > buf_ram_out;
  sc_signal<sc_uint<3> >  buf_ram_addr;
  sc_signal<bool>	  buf_ram_we;
   
  // Multiplier
  sc_signal<sc_uint<32> > mul_a;
  sc_signal<sc_uint<32> > mul_b;
  sc_signal<sc_uint<32> > mul_c;
    
  memoria_rom *buf0;
  memoria_ram *buf1;
  multiplier_28_32 *mul0;
 
  // Maquina de estados
  void fsm();


  //typedef imdctwindow SC_CURRENT_USER_MODULE;
  //imdctwindow::imdctwindow(sc_module_name mod_imdctwindow) : sc_module(mod_imdctwindow)
  SC_CTOR( imdctwindow )
    {
      buf0 = new memoria_rom("buf0");
      buf0->clk(clk);
      buf0->out_data(buf_rom_out);
      buf0->addr(buf_rom_addr);

      buf1 = new memoria_ram("buf1");
      buf1->clk(clk);
      buf1->in_data(buf_ram_in);
      buf1->out_data(buf_ram_out);
      buf1->we(buf_ram_we);
      buf1->addr(buf_ram_addr);

      mul0 = new multiplier_28_32("mul0");
      mul0->a(mul_a);
      mul0->b(mul_b);
      mul0->c(mul_c);

      SC_METHOD(fsm);
      sensitive_pos << clk;
    }

};
#endif
