// +FHDR------------------------------------------------------------------------ 
// Copyright (c) Brazil-IP / CT-INFO  -  Fenix
// ----------------------------------------------------------------------------- 
// FILE NAME : dcdct.h
// UNIVERSITY : Universidade Federal de Campina Grande - UFCG
// AUTHOR : Roberto Medeiros de Faria
// AUTHOR S EMAIL : rfaria@dsc.ufcg.edu.br
// ----------------------------------------------------------------------------- 
// RELEASE HISTORY 
// VERSION DATE AUTHOR DESCRIPTION 
// 1.0 YYYY-MM-DD name 
// ----------------------------------------------------------------------------- 
// KEYWORDS : General file searching keywords, leave blank if none. 
// ----------------------------------------------------------------------------- 
// PURPOSE : This is a part of Fenix plataform that implements the run length
//           decoder of a MPEG-4 controler.
// ----------------------------------------------------------------------------- 
// MODULE PARAMETERS 
// PARAM NAME RANGE : DESCRIPTION : DEFAULT : UNITS 
// ----------------------------------------------------------------------------- 
// OCP-IP PARAMETERS
// reset, datahandshake
// ----------------------------------------------------------------------------- 
// REUSE ISSUES 
// Reset Strategy : 
// Clock Domains : 
// Critical Timing : 
// Test Features : 
// Asynchronous I/F : 
// Scan Methodology : 
// Instantiations : 
// Synthesizable (y/n) : 
// Other : 
// -FHDR------------------------------------------------------------------------


#include "mpeg_def.h"
#include "systemc.h"

SC_MODULE (dcdct) {

  sc_in <bool> clk;   // clock signal
  sc_in <bool> reset; // reset signal

  /* input interface from Bitstream */
  sc_in <sc_uint<COEFFS_PER_BLOCK_SIZE> > run; // number of zeros coefficients before non zero
  sc_in <sc_int<COEFF_SIZE> > level;   // mgnitude of coefficient
  sc_in <bool> sign;                   // coefficient in sign
  sc_in <bool> last;                   // last coefficient non zero or not    
  sc_in <bool> in_valid;               // may receive the three values
  sc_out <bool> in_ready;              // ready to receive the three values
 
  /* output interface to Inverse Scan */
  sc_out <sc_int<COEFF_SIZE> > qfs;  // coeficcient to be sended
  sc_out <bool> out_qfs_valid;          // ready to send the coefficient
  sc_in <bool> out_qfs_ready;           // may send the coefficient

  /* internal signals*/



  typedef enum {IDLE = 0, GENERATE_ZEROS_BEFORE = 1, SEND_ZEROS_BEFORE = 2, SEND_LEVEL = 3, GENERATE_ZEROS_AFTER = 4, FINISH = 5 } state;
  sc_signal <sc_uint<3> > qfs_state;                      // state of state machine
  sc_signal <sc_uint<COEFFS_PER_BLOCK_SIZE> > count_zero;   // conter to generate zeros coefficients
  sc_signal <sc_uint<COEFFS_PER_BLOCK_SIZE> > num_coef_gen; // number of coefficients generated yet
  sc_signal <bool> last_;
  sc_signal <sc_int<COEFF_SIZE> > level_;

  void decoder();

  SC_CTOR(dcdct) {  
    SC_METHOD(decoder);         
    sensitive_pos << clk; 
  }

}; // dcdct
